import getopt
from typing import Any, NoReturn, Union
import socket
import sys
from wsgiref.simple_server import WSGIServer, WSGIRequestHandler
from .gunicorn_utils import import_app, parse_address, is_ipv6


class WSGIServer6(WSGIServer):
    address_family = socket.AF_INET6


USAGE: str = """\
Usage: {name:s} [-b ADDRESS | --bind ADDRESS] APPLICATION

ADDRESS:
    A string in HOST:PORT format, where HOST is a valid IP address and PORT is a
    valid TCP port.

APPLICATION:
    A WSGI application import path. The format is the same that Gunicorn uses.
    Read more at <https://docs.gunicorn.org/en/stable/settings.html#wsgi-app>.
"""


def print_usage(file=sys.stdout) -> None:
    print(USAGE.format(name=sys.argv[0]), file=file, end="")


def die_with_usage(msg:str=None) -> NoReturn:
    if msg is not None:
        print(msg, file=sys.stderr)
    print_usage(sys.stderr)
    sys.exit(1)


def make_server(host: str, port: int, app: Any) -> Union[WSGIServer, WSGIServer6]:
    res = socket.getaddrinfo(
        host,
        port,
        type=socket.SOCK_STREAM,
        proto=socket.getprotobyname("tcp")
    )
    if any(e[0] is socket.AF_INET6 for e in res):
        family, _, _, _, sockaddr = next(e for e in res if e[0] is socket.AF_INET6)
    else:
        family, _, _, _, sockaddr = res[0]
    if family is socket.AF_INET:
        server_class = WSGIServer
    elif family is socket.AF_INET6:
        server_class = WSGIServer6
    else:
        raise ValueError("Unsupported family: {!s:s}".format(family))
    server = server_class(sockaddr, WSGIRequestHandler, bind_and_activate=False)
    server.set_app(app)
    try:
        server.server_bind()
        server.server_activate()
    except:
        server.server_close()
        raise
    return server


def main():
    if len(sys.argv) == 0:
        print_usage()
        sys.exit(0)
    try:
        optlist, args = getopt.getopt(sys.argv[1:], "b:", ["bind"])
    except getopt.GetOptError as exc:
        die_with_usage(str(exc))
    if len(args) < 1:
        die_with_usage()
    app_uri = args[0]
    bind_address = None
    for option, value in optlist:
        if option in ("-b", "--bind"):
            bind_address = value
        else:
            assert False, "DEAD CODE"
    if bind_address is None:
        die_with_usage()
    host, port = parse_address(bind_address)
    app = import_app(app_uri)
    with make_server(host, port, app) as httpd:
        print("Listening on http://{:s}:{:d}/".format(("[" + host + "]") if is_ipv6(host) else host, port))
        httpd.serve_forever()



# This code was shamelessly stolen from the great Gunicorn project.



# 2009-2018 (c) Benoît Chesneau <benoitc@e-engura.org>
# 2009-2015 (c) Paul J. Davis <paul.joseph.davis@gmail.com>

# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:

# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.


import ast
import importlib
import logging
import os
import socket
import sys
import traceback


class AppImportError(Exception):
    """ Exception raised when loading an application """


def _called_with_wrong_args(f):
    """Check whether calling a function raised a ``TypeError`` because
    the call failed or because something in the function raised the
    error.

    :param f: The function that was called.
    :return: ``True`` if the call failed.
    """
    tb = sys.exc_info()[2]

    try:
        while tb is not None:
            if tb.tb_frame.f_code is f.__code__:
                # In the function, it was called successfully.
                return False

            tb = tb.tb_next

        # Didn't reach the function.
        return True
    finally:
        # Delete tb to break a circular reference in Python 2.
        # https://docs.python.org/2/library/sys.html#sys.exc_info
        del tb


def import_app(module):
    parts = module.split(":", 1)
    if len(parts) == 1:
        obj = "application"
    else:
        module, obj = parts[0], parts[1]

    try:
        mod = importlib.import_module(module)
    except ImportError:
        if module.endswith(".py") and os.path.exists(module):
            msg = "Failed to find application, did you mean '%s:%s'?"
            raise ImportError(msg % (module.rsplit(".", 1)[0], obj))
        raise

    # Parse obj as a single expression to determine if it's a valid
    # attribute name or function call.
    try:
        expression = ast.parse(obj, mode="eval").body
    except SyntaxError:
        raise AppImportError(
            "Failed to parse %r as an attribute name or function call." % obj
        )

    if isinstance(expression, ast.Name):
        name = expression.id
        args = kwargs = None
    elif isinstance(expression, ast.Call):
        # Ensure the function name is an attribute name only.
        if not isinstance(expression.func, ast.Name):
            raise AppImportError("Function reference must be a simple name: %r" % obj)

        name = expression.func.id

        # Parse the positional and keyword arguments as literals.
        try:
            args = [ast.literal_eval(arg) for arg in expression.args]
            kwargs = {kw.arg: ast.literal_eval(kw.value) for kw in expression.keywords}
        except ValueError:
            # literal_eval gives cryptic error messages, show a generic
            # message with the full expression instead.
            raise AppImportError(
                "Failed to parse arguments as literal values: %r" % obj
            )
    else:
        raise AppImportError(
            "Failed to parse %r as an attribute name or function call." % obj
        )

    is_debug = logging.root.level == logging.DEBUG
    try:
        app = getattr(mod, name)
    except AttributeError:
        if is_debug:
            traceback.print_exception(*sys.exc_info())
        raise AppImportError("Failed to find attribute %r in %r." % (name, module))

    # If the expression was a function call, call the retrieved object
    # to get the real application.
    if args is not None:
        try:
            app = app(*args, **kwargs)
        except TypeError as e:
            # If the TypeError was due to bad arguments to the factory
            # function, show Python's nice error message without a
            # traceback.
            if _called_with_wrong_args(app):
                raise AppImportError(
                    "".join(traceback.format_exception_only(TypeError, e)).strip()
                )

            # Otherwise it was raised from within the function, show the
            # full traceback.
            raise

    if app is None:
        raise AppImportError("Failed to find application object: %r" % obj)

    if not callable(app):
        raise AppImportError("Application object must be callable.")
    return app


def parse_address(netloc, default_port='8000'):
    if netloc.startswith("tcp://"):
        netloc = netloc.split("tcp://")[1]
    host, port = netloc, default_port

    if '[' in netloc and ']' in netloc:
        host = netloc.split(']')[0][1:]
        port = (netloc.split(']:') + [default_port])[1]
    elif ':' in netloc:
        host, port = (netloc.split(':') + [default_port])[:2]
    elif netloc == "":
        host, port = "0.0.0.0", default_port

    try:
        port = int(port)
    except ValueError:
        raise RuntimeError("%r is not a valid port number." % port)

    return host.lower(), port


def is_ipv6(addr):
    try:
        socket.inet_pton(socket.AF_INET6, addr)
    except socket.error:  # not a valid address
        return False
    except ValueError:  # ipv6 not supported on this platform
        return False
    return True
